/*
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of Ant Kutschera's blog.
 * 
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.test.services;

import java.util.Date;
import java.util.List;

import javax.ejb.EJB;
import javax.ejb.Schedule;
import javax.ejb.Stateless;
import javax.ejb.Timer;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;

import uk.co.maxant.test.CommsStatus;
import uk.co.maxant.test.data.CommsLog;
import uk.co.maxant.test.ws.PricingSystemException_Exception;

@Stateless
@TransactionManagement(TransactionManagementType.CONTAINER)
public class CleanupBeanTimer {

	@PersistenceContext(unitName="SevenJPA")
	private EntityManager em;
	
	@EJB
	private PartnerBeanLocal partner;

    public CleanupBeanTimer() {
    }
	
	/**
	 * if there are any old (> 20 second old) offers, cancel all CLOGs
	 * for the entire purchase.
	 * 
	 * in reality, there could also be failures after calling the web services,
	 * so ideally we would ensure that any purchaseUuids that do not exist in the purchase
	 * table, that they are all cancelled
	 */
	@SuppressWarnings("unused")
	@Schedule(second="30", minute="*/1", hour="4-3", dayOfWeek="Mon-Sun", dayOfMonth="*", month="*", year="*", info="CancelFailedPurchasesTimer")
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	private void cancelFailedPurchases(final Timer t) {
		long start = System.currentTimeMillis();
		System.out.println("@CancelFailedPurchases called at: " + new java.util.Date());
		
		TypedQuery<CommsLog> q = em.createQuery(
				"select clog " +
				"from CommsLog clog " +
				"where clog.status = :status and " +
				"clog.whenSent < :when", CommsLog.class);
		q.setParameter("status", CommsStatus.OFFERED.toString());
		q.setParameter("when", new Date(System.currentTimeMillis() - (1000L*20)));
		List<CommsLog> incompleteClogs = q.getResultList();
		
		for(CommsLog incompleteClog : incompleteClogs){
			//we need to cancel all offers in this business case as we couldnt complete it.
			q = em.createQuery(
					"select clog " +
					"from CommsLog clog " +
					"where clog.purchaseUuid = :p", CommsLog.class);
			q.setParameter("p", incompleteClog.getPurchaseUuid());
			List<CommsLog> clogsInBC = q.getResultList();

			for(CommsLog c : clogsInBC){
				try {
					partner.cancel(c.getOfferId());
					System.out.println("cancelled incomplete clog: " + c.getUid());
					c.setStatus(CommsStatus.CANCELLED.toString());
				} catch (PricingSystemException_Exception e) {
					System.err.println("FAILED TO CANCEL OFFER FROM CLOG " + c.getUid() + ". PLEASE DO IT MANUALLY. " + e.getMessage());
					c.setStatus(CommsStatus.MANUAL.toString());
				}
			}
		}
		System.out.println("@CancelFailedPurchases finished in " + (System.currentTimeMillis() - start) + " ms");
	}


	 /* 
	 * TODO if we fail after purchasing, our recovery still needs to cancel everything, so it needs 
	 * to check the message table which is only committed in the outer transaction. ie add the business
	 * case to the message table.
	 */

}