/*
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of Ant Kutschera's blog.
 * 
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.test.services;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.annotation.security.DeclareRoles;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.EJBException;
import javax.ejb.SessionContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.interceptor.Interceptors;
import javax.jms.JMSException;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;

import uk.co.maxant.test.BusinessException;
import uk.co.maxant.test.Constants;
import uk.co.maxant.test.Roles;
import uk.co.maxant.test.data.Product;
import uk.co.maxant.test.data.Purchase;
import uk.co.maxant.test.ws.Offer;
import uk.co.maxant.test.ws.PricingSystemException;
import uk.co.maxant.test.ws.PricingSystemException_Exception;

/**
 * gets offers for a trip to the zoo, for the following items:
 * 
 * 1) entrance ticket
 * 2) bag of elephant treats for feeding the elephants
 * 3) "My Parents visited the Zoo and all I got was 
 *    this lousy T-Shirt" T-Shirt
 * 
 * the offers take a while to be returned, because the partner
 * system has some complex logic and needs to interact with 
 * for example reservation systems, stock control systems, etc.
 * 
 * so to get the offers for these items, we make the three calls
 * in parallel.
 * 
 * afterwards, we purchase all items.
 * 
 * in the case of an error (which can be simulated using a command)
 * we simply return an error to the user. there is then a scheduled 
 * task ({@link CleanupBeanTimer#cancelFailedPurchases(Timer)}
 * which ensures all items are cancelled.
 * 
 * once the items have been purchased, we send a confirmation email
 * to the logged in user, but asynchronously (for two reasons - 
 * it can be made transactional by doing this, as well as it saves 
 * time coz mail is slow). finally we write this transaction into
 * the purchases table.
 */
@DeclareRoles({Roles.REGISTERED})
@Stateless
@TransactionManagement(TransactionManagementType.CONTAINER)
public class OrchestrationBean implements OrchestrationBeanLocal {

	@EJB
	private PartnerBeanLocal partner;
	
	@Resource(mappedName = "jms/destinationResource")
	private Queue q;

	@Resource(mappedName = "jmsConnectionPool")
	private QueueConnectionFactory qcf;

	@Resource 
	private SessionContext sessionContext;

	@PersistenceContext(unitName="SevenJPA")
	private EntityManager em;
	
	private QueueConnection qc = null;
	private QueueSession qs = null;
	private QueueSender qSender = null;
	
	@PostConstruct
	protected void init(){
		try {
			qc = qcf.createQueueConnection();
			qs = qc.createQueueSession(true, Session.SESSION_TRANSACTED);
			qSender = qs.createSender(q);
		} catch (JMSException e) {
			throw new RuntimeException("unable to create ejb", e);
		}
	}
	
	@PreDestroy
	protected void close(){
		if (qc != null) {
			try {
				qc.close();
			} catch (JMSException e) {
				throw new RuntimeException("unable to destroy ejb", e);
			}
		}
	}

	/**
	 * Default constructor.
	 */
	public OrchestrationBean() {
	}

	@Override
	@RolesAllowed({Roles.REGISTERED})
	@TransactionAttribute(TransactionAttributeType.REQUIRED)
	@Interceptors({PerformanceInterceptor.class})
	public String execute(String command) throws BusinessException {

		String ret = "hello " + sessionContext.getCallerPrincipal() + ", you gave command: " + command;
		try {
			UUID purchaseId = UUID.randomUUID();

			//go get offers for all the items and purchase them.
			purchaseItems(purchaseId, command);

			//write something to the database
			addPurchaseToDB(purchaseId);
			
			//send email
			sendAsyncEmail(purchaseId);
			
			//handle command for error simulation
			handleCommand(command);
			
		} catch (JMSException e) {
			ret = "FAILED JMS: " + e.getMessage();
		} catch (EJBException e) {
			ret = "FAILED EJB: ";
			if(e.getCause() != null){
				ret += e.getCause().getMessage();
			}else{
				ret += e.getMessage();
			}
		} catch (ExecutionException e) {
			ret = "FAILED EXEC: ";
			if(e.getCause() != null){
				if(e.getCause().getCause() != null){
					ret += e.getCause().getCause().getMessage();
				}else{
					ret += e.getCause().getMessage();
				}
			}else{
				ret += e.getMessage();
			}
		} catch (Exception e) {
			ret = "FAILED: " + e.getMessage();
		}

		//since we catch Exception above, theres no point in throwing a system exception within that
		//try/catch, since its caught, and results in a commit. so throw it somewhere where the container
		//will catch it!
		if(command.equals("systemException")){
			throw new RuntimeException("simulating a system exception");
		}

		return ret;
	}

	/**
	 * gets offers for each item and purchases them.
	 * 
	 * @param purchaseId the UUID for this purchase
	 * @param command error simulation command
	 * @throws PricingSystemException_Exception if pricing partner system has an exception
	 * @throws BusinessException if offers for all items could not be retrieved
	 * @throws InterruptedException if async call to partner fails for some reason
	 * @throws ExecutionException if async call to partner has an exception
	 */
	private void purchaseItems(UUID purchaseId, String command) throws PricingSystemException_Exception, BusinessException, InterruptedException, ExecutionException {
		
		long start = System.currentTimeMillis();

		//go get offers for all three items in parallel
		List<Future<Offer>> offers = getOffersInParallel(command);

		//now wait for all offers to come back
		new AsyncTracker<Offer>().waitForCompletion(offers);

		logOfferResults(start, offers);

		purchaseOffers(command, offers);
		
	}

	/**
	 * goes and gets offers for the items, asynchronously.
	 * @param command command to simulate errors
	 * @return the offers
	 */
	private List<Future<Offer>> getOffersInParallel(String command) {
		List<Future<Offer>> results = new ArrayList<Future<Offer>>();
		UUID businessUuid = UUID.randomUUID();
		for(int i = 0; i < 3; i++){
			//the last one will get an invalid name if required
			Product p = new Product();
			switch (i) {
				case 0:
					p.setId("ENTRANCE-TICKET");
					break;
				case 1:
					p.setId("TASTY-ELEPHANT-TREAT");
					break;
				case 2:
					if(command.equals("validationError")){
						p.setId("T-SHIRT with some really long product ID that will fail validation");
					}else{
						p.setId("T-SHIRT");
					}
					break;
			}
			results.add(partner.getOffer(p, businessUuid ));
		}
		return results;
	}

	/**
	 * logs how long it took to get results, and what they were.
	 */
	private void logOfferResults(long start, List<Future<Offer>> results)
			throws InterruptedException, ExecutionException {
		String result = "";
		for (Future<Offer> r : results) {
			result += r.get().getProductId()
					+ ": "
					+ (r.isDone() ? "done" : (r.isCancelled() ? "cancelled"
							: "unknown")) + "; ";
		}
		System.out.println("Finished async long tasks (" + result + ") in "
				+ (System.currentTimeMillis() - start) + "ms");
	}
	
	/**
	 * purchases the given offers from the partner system.
	 * @param command used to simulate errors
	 * @param offers the offers to purchase
	 */
	private void purchaseOffers(String command, List<Future<Offer>> results)
			throws InterruptedException, ExecutionException, BusinessException,
			PricingSystemException_Exception {
		
		// if we dont have three, we cannot continue!
		List<Offer> offers = new ArrayList<Offer>();
		for (Future<Offer> r : results) {
			if (r.isDone()) {
				offers.add(r.get());
			} else {
				throw new BusinessException(
						"Cannot continue without three successful offers!");
			}
		}

		// buy the offers
		NumberFormat nf = NumberFormat.getCurrencyInstance(Constants.LOCALE);
		for (Offer o : offers) {
			System.out.println("Purchasing offer " + o.getOfferId() + ", "
					+ o.getProductId() + ", price=" + nf.format(o.getPrice()));
			if (command.equals("purchase")
					&& o.getProductId().equals("T-SHIRT")) {
				throw new PricingSystemException_Exception(
						"simulating an exception in the partner system",
						new PricingSystemException());
			}
			partner.purchase(o.getOfferId());
		}
	}

	/**
	 * writes a purchase to the database.
	 */
	private void addPurchaseToDB(UUID purchaseUuid) {
		
		Purchase p = new Purchase();
    	p.setPurchaseUuid(purchaseUuid.toString());
		p.setCustomer(sessionContext.getCallerPrincipal().getName());
		p.setWhenBought(new Date());
		em.persist(p);
	}

	/**
	 * sends a JMS message in order to send an email confirmation
	 */
	private void sendAsyncEmail(UUID purchaseUuid) throws JMSException {
		TextMessage msg = qs.createTextMessage();
		msg.setText("Thanks for purchasing the special Zoo Pack, comprising entrance, elephant " +
				"treats and T-shirt. Your e-ticket number is: " + purchaseUuid.toString() + ". Please " +
						"print this email and show it with your credit card at the Zoo when you arrive." +
						"Yours sincerely, TeamZoo");
		msg.setStringProperty(MailSenderBean.TO, sessionContext.getCallerPrincipal().getName());
		qSender.send(msg);
	}

	/**
	 * handle command for error simulation
	 * @throws BusinessException 
	 */
	private void handleCommand(String command) throws BusinessException {
		if(command.equals("rollbackOnly")){
			sessionContext.setRollbackOnly();
		}else if(command.equals("businessExceptionWithRollback")){
			sessionContext.setRollbackOnly();
			throw new BusinessException("simulating a business exception (called setRollbackOnly too!)");
		}else if(command.equals("businessExceptionWithoutRollback")){
			throw new BusinessException("simulating a business exception (didnt called setRollbackOnly!!!)");
		}
	}

}
