/*
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of Ant Kutschera's blog.
 * 
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.test.services;

import java.sql.Timestamp;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.Future;

import javax.annotation.Resource;
import javax.annotation.security.DeclareRoles;
import javax.annotation.security.RolesAllowed;
import javax.ejb.AsyncResult;
import javax.ejb.Asynchronous;
import javax.ejb.SessionContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.interceptor.Interceptors;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;

import uk.co.maxant.test.CommsStatus;
import uk.co.maxant.test.Roles;
import uk.co.maxant.test.data.CommsLog;
import uk.co.maxant.test.data.Product;
import uk.co.maxant.test.ws.Offer;
import uk.co.maxant.test.ws.PricingSystem;
import uk.co.maxant.test.ws.PricingSystemException_Exception;

/**
 * calls a web service, so may take a while to get back...
 */
@Stateless
@DeclareRoles({Roles.REGISTERED})
@TransactionManagement(TransactionManagementType.CONTAINER)
public class PartnerBean implements PartnerBeanLocal {

	//we could do the following, but it hardly helps
	//if the service is secure or over ssl, we would need to 
	//have the wsdl file locally too (ie url doesnt work)
	//@WebServiceRef(wsdlLocation="http://localhost:8084/SevenWS/PricingSystemService?wsdl")
	//private PricingSystemService pricingSystem;
	
	@Resource
	private SessionContext sessionContext;
	
	@PersistenceContext(unitName="SevenJPA")
	private EntityManager em;
	
    public PartnerBean() {
    }
    
    /**
     * note that async methods implicitly start new transactions! 
     * see ejb 3.1 spec, chapter 4.5.3.
     */
    @Override
    @Asynchronous
	@RolesAllowed({Roles.REGISTERED})
	@Interceptors({ValidationInterceptor.class})
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
    public Future<Offer> getOffer(Product product, UUID purchaseId) {

    	long start = System.currentTimeMillis();

        //instantiate and call the WS
    	PricingSystem proxy = PricingSystemFactory.createPricingSystem();
    	Offer o = proxy.getPrice(product.getId());
    	
    	//write status into database
    	//this is useful, as it allows us to implement a quasi-two-phase-commit.
    	//ie we log that we called them using this
    	CommsLog clog = new CommsLog();
    	clog.setPurchaseUuid(purchaseId.toString());
    	clog.setOfferId(o.getOfferId());
    	clog.setPrice(o.getPrice());
    	clog.setProductId(o.getProductId());
    	clog.setSender(sessionContext.getCallerPrincipal().getName());
    	clog.setStatus(CommsStatus.OFFERED.toString());
    	clog.setTimeTaken((int)(System.currentTimeMillis()-start));
    	clog.setWhenSent(new Timestamp(start));
    	em.persist(clog);
    	
		return new AsyncResult<Offer>(o);
    }
    
    /**
     * use REQUIRES_NEW so that we commit the comms log regardless of errors
     */
    @Override
	@RolesAllowed({Roles.REGISTERED})
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
    public void purchase(String offerId) throws PricingSystemException_Exception {
        //instantiate and call the WS
    	PricingSystem proxy = PricingSystemFactory.createPricingSystem();
    	proxy.purchaseOffer(offerId);
    	
    	updateClog(offerId, CommsStatus.BOUGHT);
    }

    /**
     * use REQUIRES_NEW so that we commit the comms log regardless of errors
     * @throws PricingSystemException_Exception 
     */
    @Override
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
    public void cancel(String offerId) throws PricingSystemException_Exception {
        //instantiate and call the WS
    	PricingSystem proxy = PricingSystemFactory.createPricingSystem();
    	proxy.cancel(offerId);
    	
    	updateClog(offerId, CommsStatus.CANCELLED);
    }

	private void updateClog(String offerId, CommsStatus status) {
		//now update the comms log
		TypedQuery<CommsLog> q = em.createQuery(
				"select clog " +
				"from CommsLog clog " +
				"where clog.offerId = :offerId", CommsLog.class);
		q.setParameter("offerId", offerId);
		
		List<CommsLog> clogs = q.getResultList();
		if(clogs.isEmpty()){
			throw new IllegalStateException("failed to find offer " + offerId + " in CLOG.");
		}else if(clogs.size() > 1){
			throw new IllegalStateException("too many CLOGS for offer " + offerId);
		}else{
			clogs.get(0).setStatus(status.toString());
		}
	}

}
