/*
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of Ant Kutschera's blog.
 * 
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.test.ws;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.jws.WebMethod;
import javax.jws.WebService;

/**
 * a service for getting an offer for a product and subsequently purchasing that offer.
 * if the bought offer needs to be cancelled, it can be.
 */
@WebService
public class PricingSystem {

	private final Random random = new Random();
	
	private static final Map<String, Offer> offers = new HashMap<String, Offer>();
	
	/**
	 * @return a list of sorted offers, ordered by creation date.
	 */
	public static List<Offer> getOffers() {
		List<Offer> sorted = new ArrayList<Offer>(offers.values());
		Collections.sort(sorted, new Comparator<Offer>() {
			@Override
			public int compare(Offer o1, Offer o2) {
				return o1.getCreated().compareTo(o2.getCreated());
			}
		});

		return sorted;
	}

	enum Status {
		OFFERED,
		SOLD,
		CANCELLED;
	}

	@WebMethod
	public Offer getPrice(String productId){

		long start = System.currentTimeMillis();
		
    	//simulate a long running task by waiting for roughly a second
		try {
			Thread.sleep(900L + random.nextInt(200));
		} catch (InterruptedException e) {
			//damn it, but not tragic
		}
		System.out.println("done long running task for " + productId + 
				". Started: " + new Date(start) + ", completed in " + 
				(System.currentTimeMillis()-start) + "ms");

		Offer offer = new Offer();
		offer.setPrice(10 * random.nextDouble());
		offer.setProductId(productId);

		offers.put(offer.getOfferId(), offer);
		
		return offer;
	}
	
	@WebMethod
	public void purchaseOffer(String offerId) throws PricingSystemException {
		if(offers.containsKey(offerId)){
			Offer o = offers.get(offerId);
			if(o.status == Status.OFFERED){
				o.status = Status.SOLD;
			}else{
				throw new PricingSystemException("cannot purchase an offer in status " + o.status);
			}
		}else{
			throw new PricingSystemException("unknown offer " + offerId);
		}
	}
	
	@WebMethod
	public void cancel(String offerId) throws PricingSystemException {
		if(offers.containsKey(offerId)){
			Offer o = offers.get(offerId);
			if(o.status == Status.SOLD){
				o.status = Status.CANCELLED;
			}else if(o.status == Status.OFFERED){
				o.status = Status.CANCELLED;
			}else{
				throw new PricingSystemException("cannot cancel offer in status " + o.status);
			}
		}else{
			throw new PricingSystemException("unknown offer " + offerId);
		}
	}
	
}
